"""

File:    PbxConfigurator.py
Author:  Jonathan Maurice <jmaurice@sangoma.com>
Date:    4/5/2011 10:51:04 AM
 
Description: This create a gateway configuration base on an xml file provided
             by a PBX such as 3CX.
 
Copyright (C) 2003-2011 Sangoma Technologies. 
All Rights Reserved.

This code is Sangoma Technologies Confidential Property.
Use of and Access to this code is covered by a previously executed
Non-disclosure Agreement between Sangoma Technologies and the Recipient.
This code is being supplied for evaluation purposes only and is not to be
used for any other purpose.  

"""

__version__ = "1.0"

import os
import shutil

from paraxip.infra.Assert import *
from paraxip.infra.PyLoggingFunctions import  *
from netborderPyInfra import TraceScope
from paraxip.pbx_configurator.PbxConfigurationInfo import PbxConfigurationInfo
from paraxip.pbx_configurator.PbxConfigManager import PbxConfigManager
from paraxip.pbx_configurator.PbxConfigJobRunner import PbxConfigJobRunner
from paraxip.infra.PyJob import JobStatus


fsLogger = Logger("netborder.pbxConfigurator")


class PbxConfigurator(object):
  """Create gateway configuration files based on a PBX xml file"""

  PBXCFG_ELEM_NAME = 'pbxConfiguration'

  def __init__(self, pbxFilePath, configDirPath):
    t = TraceScope(fsLogger, "PbxConfigurator.__init__")
    super(PbxConfigurator, self).__init__()
    self.pbxFilePath = pbxFilePath
    self.configDirPath = configDirPath
    self.pbxConfigInfo = None
    self.pbxJobStatus = None
    self.pbxConfigManager = None
    self.pbxConfigJobRunner = None

  #-----------------------------------------------------------------------------#
  # @return 0 : everything is fine
  # @return 1 : application error. See log file for more details
  # @return 2 : invalid configuration file
  # @return 3 : invalid configuration version
  def initialize(self):
    """Initialize the whole process."""
    t = TraceScope(fsLogger, "PbxConfigurator.initialize")

    try:
      self.pbxConfigInfo = PbxConfigurationInfo(pbxFilePath=self.pbxFilePath)
    except Exception, e:
      log_fatal(fsLogger, str(e))
      return 2

    #Parse and validate the pbx configuration file.
    if not self.validatePbxFile():
      log_fatal(fsLogger, "Pbx configuration file is invalid")
      return 3

    #Initialize required PbxConfigurator objects.
    self.pbxJobStatus = JobStatus()
    pbxFileExt = os.path.splitext(self.pbxFilePath)[1]
    replaceDict = {'na_me':'name'}
    self.pbxConfigManager = PbxConfigManager(self.pbxFilePath,
                                             self.pbxConfigInfo,
                                             self.configDirPath,
                                             replaceStrDict=replaceDict,
                                             replaceFileExtension=pbxFileExt)

    if self.pbxConfigManager is None:
      log_fatal(fsLogger, "Failed to create PbxConfigManager")
      return 1

    if not self.pbxConfigManager.initialize():
      log_fatal(fsLogger, "Failed to initialize PbxConfigManager")
      return 1

    self.pbxConfigJobRunner = PbxConfigJobRunner(self.pbxJobStatus,
                                                 self.pbxConfigManager)

    return 0


  def validatePbxFile(self):
    """ Validate that the version in the pbx file is the same as this module"""
    t = TraceScope(fsLogger, "PbxConfigurator.validatePbxFile")

    #Validate version
    if __version__ != self.pbxConfigInfo.getVersion():
      log_error(fsLogger, "Configuration file version '%s' cannot be use with "\
                "this version of the pbx-configurator '%s'" % 
                (self.pbxConfigInfo.getVersion(), __version__)
      )
      return False

    return True


  def run(self):
    """Create gateway configuration files."""
    t = TraceScope(fsLogger, "PbxConfigurator.run")

    log_info(fsLogger, "Creating %s configuration from %s" %
            (self.pbxConfigInfo.getAppName(), self.pbxFilePath)
    )
    #Add all jobs.
    self.pbxConfigJobRunner.configure()

    if not self.pbxConfigJobRunner.run():
      log_error(fsLogger, "Failed to complete gateway configuration")
      return False
    
    workingDirPath = self.pbxConfigManager.getWorkingDirectory()
    if os.path.exists(workingDirPath):
      log_debug(fsLogger, "Deleting working directory '%s'" % workingDirPath)
      try:
        shutil.rmtree(workingDirPath)
      except (OSError, IOError), e:
        log_warn(fsLogger, "Unable to remove directory '%s'. Error=%s" %
                (workingDirPath, e)
        )
    
    return True


def main(inputCfgFile, outputDir):
  pbxCfg = PbxConfigurator(inputCfgFile, outputDir)
  rc = pbxCfg.initialize()
  if not rc == 0:
    log_error(fsLogger, "Initialization failed with code: %s" % rc)
    return rc
  if not pbxCfg.run():
    log_error(fsLogger, "PBX Configurator process failed")
    return 1

  return 0

###############################################################################
# Runtest
if __name__ == '__main__':
  
  fsLogger.setLogLevel(LogLevel.debug)
  Logger('netborder.infra.PyJob').setLogLevel(LogLevel.debug)
  Logger('netborder.smartUpdate').setLogLevel(LogLevel.debug)
  
  rc = main('test-input/3cx-a200-2fxo-2fxs.pbx', 'test-output/config-dir')
  if rc == 0:
    log_info(fsLogger, "Runtest passed")
    print("Passed rc=%s" % str(rc))
  else:
    log_error(fsLogger, "Runtest failed")
    print("Failed rc=%s" % str(rc))

## !!MIRROR FILE, DO NOT MODIFY!!
