"""

File: TempDirManager.py
Author: Jonathan Maurice
Date: 14/02/2011
Description: Directory manager of the smart update. This class is used to
alternate between 2 temporary directories from which the converters 
operation take place.

Copyright (C) 2003-2011 Sangoma Technologies.
All Rights Reserved.

"""

from paraxip.infra.Assert import *
from paraxip.infra.PyLoggingFunctions import  *
from netborderPyInfra import TraceScope
from netborderPyInfra import PyGlobalConfig as globalConfig

import os
import re
import shutil
import fileinput

fsLogger = Logger("netborder.smartUpdate.TempDirManager")


class TempDirManager(object):
  """Manage a set of 2 working directories which are used by the smart update
  converters to make their operation between them.
  Args:
    initialConfigDir: Path to the initial config (extracted in zip file)
    finalConfigDir: Path to the final config (end of smart update)
    replaceStrDict: Dictionnary of string to replace of this format
      {'stringToFind':'stringToReplaceWith'}
    replaceFileExtension: A string or a list of string of extension in wich
      the replace will be done"""

  INIT_MODE, TEMP_MODE, FINAL_MODE =  range(3)

  def __init__(self, initialConfigDir, finalConfigDir, replaceStrDict=None,
               replaceFileExtension=None):
    t = TraceScope(fsLogger, "TempDirManager.__init__")
    super(TempDirManager, self).__init__()
    self._initCfgDir = initialConfigDir
    self._finalCfgDir = finalConfigDir
    self._replaceDict = replaceStrDict
    
    if replaceFileExtension is None:
      self._replaceFileExtension = list()
    elif isinstance(replaceFileExtension, list):
      self._replaceFileExtension = replaceFileExtension
    else:
      self._replaceFileExtension = list()
      self._replaceFileExtension.append(replaceFileExtension)

    self._srcDir = None
    self._dstDir = None
    self._mode = None


  def initialize(self):
    """Make sure temporary directories are deleted and set mode to Initial"""
    t = TraceScope(fsLogger, "TempDirManager.initialize")

    self._workingDir = \
        globalConfig.getParameter("netborder.smartUpdate.Working.Directory")
    
    if self._workingDir is None:
      logAssert(fsLogger, "netborder.smartUpdate.Working.Directory as not been"\
                "set.")
      return None

    self._tmpDir1 = os.path.join(self._workingDir, "tmpDir1")
    self._tmpDir2 = os.path.join(self._workingDir, "tmpDir2")
    
    if self.cleanTempDirectories() is False:
      return False

    
    return self.setInitialMode()


  def getSrcDir(self):
    t = TraceScope(fsLogger, "TempDirManager.getSrcDir")
    return self._srcDir


  def getDstDir(self):
    """docstring for getDstDir"""
    t = TraceScope(fsLogger, "TempDirManager.getDstDir")
    return self._dstDir


  def swapDirectories(self):
    """Swap source and destination directories"""
    t = TraceScope(fsLogger, "TempDirManager.swapDirectories")
    
    if self._mode == TempDirManager.TEMP_MODE:

      bkpTmpDir = self._srcDir
      log_debug(fsLogger, "Setting source directory to '%s'" % self._dstDir)
      self._srcDir = self._dstDir

      log_debug(fsLogger, "Setting destination directory to '%s'" % bkpTmpDir)
      self._dstDir = bkpTmpDir
    else:
      log_warn(fsLogger, "Trying to swap directories while not in Temp Mode, "\
                         "directories will not be swapped.")
      return False

    return True


  def setInitialMode(self): 
    """ Set source dir to initCfgDir and dst dir to tmpDir1 """
    t = TraceScope(fsLogger, "TempDirManager.setInitialMode")

    if self._tmpDir1 is None:
      log_warn(fsLogger, "Can't set mode to initial, TempDirManager is not"\
               "initialized yet.")
      return False

    log_debug(fsLogger, "Setting mode to Initial")
    self._mode = TempDirManager.INIT_MODE

    log_debug(fsLogger, "Setting source dir to initial config dir '%s'" %
              self._initCfgDir)
    self._srcDir = self._initCfgDir

    log_debug(fsLogger, "Setting destination dir to tmpDir1 '%s'" %
              self._tmpDir1)
    self._dstDir = self._tmpDir1

    return True


  def setTempMode(self):
    """ Set srcDir to tmpDir1 and dstDir to tmpDir2"""

    t = TraceScope(fsLogger, "TempDirManager.setTempMode")
    log_debug(fsLogger, "Setting mode to Temp")
    self._mode = TempDirManager.TEMP_MODE

    log_debug(fsLogger, "Setting source dir to tmpDir1 '%s'" %
              self._tmpDir1)
    self._srcDir = self._tmpDir1

    log_debug(fsLogger, "Setting destination dir to tmpDir2 '%s'" %
              self._tmpDir2)
    self._dstDir = self._tmpDir2

    return True
  

  def setFinalMode(self):
    """ Set destination directory to _finalCfgDir """

    t = TraceScope(fsLogger, "TempDirManager.setFinalMode")
    log_debug(fsLogger, "Setting mode to Final")
    self._mode = TempDirManager.FINAL_MODE

    log_debug(fsLogger, "Setting destination dir to final config dir '%s'" %
              self._finalCfgDir)
    self._dstDir = self._finalCfgDir

    return True


  def initialConvert(self):
    """Copy files from initial config directory to temp dir 1 and replace the
    string passed in constructor in each file"""
    t = TraceScope(fsLogger, "TempDirManager.initialConvert")
    

    if self._mode != TempDirManager.INIT_MODE:
      log_warn(fsLogger, "Trying to do initial convert while not in Initial "\
                         "mode, initial convert will not be done.")
      return False

    if self._createTempDirectories() is False:
      return False

    if self._copyInitialConfig() is False:
      log_error(fsLogger, "Initial convert failed, can't copy file to"\
                          "temporary directory.")
      return False

    #Replace all string in _dstDir
    if self._replaceDict is None:
      return True
    else:
      return self._replaceStringInFiles(self._dstDir) 

  

  def cleanTempDirectories(self):
    """Remove all temporary directories used by the smart update"""
    t = TraceScope(fsLogger, "TempDirManager.cleanTempDirectories")

    if os.path.exists(self._tmpDir1):
      log_debug(fsLogger, "Deleting temporary directory '%s'" % self._tmpDir1)
      try:
        shutil.rmtree(self._tmpDir1)
      except (OSError, IOError), e:
        log_error(fsLogger, "Unable to remove directory '%s'. Error=%s" %
                  (self._tmpDir1, e))
        return False

    if os.path.exists(self._tmpDir2):
      log_debug(fsLogger, "Deleting temporary directory '%s'" % self._tmpDir2)
      try:
        shutil.rmtree(self._tmpDir2)
      except (OSError, IOError), e:
        log_error(fsLogger, "Unable to remove directory '%s'. Error=%s" %
                  (self._tmpDir2, e))
        return False

    return True


  def eraseDstDirContent(self):
    """Remove all file and subdirectories in _dst directory"""
    t = TraceScope(fsLogger, "TempDirManager.eraseDstDirContent")
    
    for currentDir, subDirs, files in os.walk(self._dstDir):
      for file in files:
        log_debug(fsLogger, "Deleting file '%s'" % file)
        try:
          os.remove(os.path.join(currentDir, file))
        except OSError, e:
          log_error(fsLogger, "Unable to delete file '%s'. Error=%s" %
                    (file, e))

      for subD in subDirs:
        log_debug(fsLogger, "Deleting directory '%s'" % subD)
        try:
          shutil.rmtree(os.path.join(currentDir, subD))
        except (OSError, IOError), e:
          log_error(fsLogger, "Unable to remove directory '%s'. Error=%s" %
                    (subD, e))

    return True


  def moveRemainingFiles(self):
    """Move files to _dstDir from _srcDir that are not in _dstDir"""
    t = TraceScope(fsLogger, "TempDirManager.moveRemainingFiles")

    srcDir = self._srcDir
    dstDir = self._dstDir

    #Make sure that srcDir and dstDir end with a delimiter (/ or \)
    if not re.search(r"[\\/]$", srcDir):
      srcDir = os.path.join(srcDir, '')

    if not re.search(r"[\\/]$", dstDir):
      dstDir = os.path.join(dstDir, '')

    #Make lists of file path that are relative to the _srcDir and _dstDir
    #To make the path relative, remove srcDir and dstDir from the file path.
    srcDirFiles = []
    dstDirFiles = []
    for file in self._getFilesInDir(srcDir):
      srcDirFiles.append(file.replace(srcDir, ''))
      
    for file in self._getFilesInDir(dstDir):
      dstDirFiles.append(file.replace(dstDir, ''))

    #Check if file in srcDir are in dstDir, if not copy the file.
    for srcFile in srcDirFiles:
      if dstDirFiles.count(srcFile):
        #File already exist
        log_debug(fsLogger, "Not moving file '%s' since destination "\
                  "directory '%s' already has this file" % (srcFile, dstDir))
        continue
      else:
        #Need to move this file
        log_debug(fsLogger, "Missing file '%s' in destination "\
                            "directory '%s'" % (srcFile, dstDir))
        if self._moveFileFromSrcToDst(srcFile) is False:
          return False

    return True


  def getInitialCfgDirPath(self):
    """Return the path to the initial config directory"""
    t = TraceScope(fsLogger, "TempDirManager.getInitialCfgDirPath")
    return self._initCfgDir

  def getPatchDataDirPath(self):
    """Return the path to the patch data directory"""
    t = TraceScope(fsLogger, "TempDirManager.getPatchDataDirPath")
    return os.path.join(self._initCfgDir, 'patch-data')

  
  def _createTempDirectories(self):
    """Create both temporary directories"""
    t = TraceScope(fsLogger, "TempDirManager._createTempDirectories")
    
    if not os.path.exists(self._tmpDir1):
      log_debug(fsLogger, "Creating temporary directory '%s'" % self._tmpDir1)
      os.makedirs(self._tmpDir1)

    if not os.path.exists(self._tmpDir2):
      log_debug(fsLogger, "Creating temporary directory '%s'" % self._tmpDir2)
      os.makedirs(self._tmpDir2)

    return True


  def _moveFileFromSrcToDst(self, fileInSrcDir):
    """Move a file in _srcDir to _dstDir, path is relative to _srcDir"""
    t = TraceScope(fsLogger, "TempDirManager._moveFileFromSrcToDst")

    srcFilePath = os.path.join(self._srcDir, fileInSrcDir)
    dstFilePath = os.path.join(self._dstDir, fileInSrcDir)
    dstDir = os.path.dirname(dstFilePath)

    if not os.path.exists(dstDir):
      #Sub directory must be created before copying the file.
      log_debug(fsLogger, "Creating subdirectories path '%s'" % dstDir)
      try:
        os.makedirs(dstDir)
      except OSError, e:
        log_error(fsLogger, "Unable to create subdirectories path '%s'. Error=%s"\
                  % (dstDir, e))
        return False


    log_debug(fsLogger, "Moving file '%s' to '%s'" % (srcFilePath, dstFilePath))
    try:
      shutil.move(srcFilePath, dstFilePath)
    except (OSError, IOError), e:
      log_error(fsLogger, "Unable to move file '%s' to '%s'. Error=%s"\
                % (srcFilePath, dstFilePath, e)) 
      return False
    
    return True


  def _copyInitialConfig(self):
    """Copy initial config directory to _dstDir"""
    t = TraceScope(fsLogger, "TempDirManager._copyInitialConfig")

    if os.path.exists(self._dstDir):
      log_debug(fsLogger, "Deleting directory '%s'" % self._dstDir)
      shutil.rmtree(self._dstDir)

    log_debug(fsLogger, "Copying directory tree '%s' to '%s'" 
                % (self._srcDir, self._dstDir))
    shutil.copytree(self._srcDir, self._dstDir)
    
    return True


  def _replaceStringInFiles(self, replaceDir):
    """Replace all occurence of a string for all files of replaceDir"""
    t = TraceScope(fsLogger, "TempDirManager._replaceStringInFiles")
    
    if self._replaceStringInFiles is None:
      log_debug(fsLogger, "Skipping replace in files since no extension were "\
                "defined")
      return True

    for file in self._getFilesInDir(replaceDir):
      if not os.path.isfile(file):
        log_warn(fsLogger, "Invalid file '%s'. Skipping string replacement." %
                 file)
        continue

      root, ext = os.path.splitext(file)
      if ext not in self._replaceFileExtension:
        log_debug(fsLogger, "Skipping replace in file '%s', replace for this "\
                  "extension is not allowed" % file)
        continue

      for line in fileinput.FileInput(file, inplace=1):
        for key in self._replaceDict.keys():
          if key in line:
            log_debug(fsLogger, "Replacing string '%s' for '%s' in file '%s'"\
                     % (key, self._replaceDict[key], file))
            line = line.replace(key, self._replaceDict[key])

        #Since FileInput as the keyword arg inplace=1 the file is moved 
        #to a backup file and standard output is directed to the input file
        sys.stdout.write(line)

    return True


  def _getFilesInDir(self, directory):
    """Return an iterator of all the files in directory and it's subdirectory.
       Path of files are relative to current working directory"""
    t = TraceScope(fsLogger, "TempDirManager._getFilesInDir")
    
    for currentDir, subDir, files in os.walk(directory):
      for currentFile in files:
        yield os.path.join(currentDir, currentFile)


###############################################################################
# Runtest
if __name__ == '__main__':

  def getFilesInDir(directory):
    for currentDir, subDir, files in os.walk(directory):
      for currentFile in files:
        yield os.path.join(currentDir, currentFile)

  def directoriesCompare(dir1, dir2):
    """Return True if directories have the same files"""

    dir1Files = list()
    dir2Files = list()

    for file in getFilesInDir(dir1):
      dir1Files.append(os.path.basename(file))
    
    for file in getFilesInDir(dir2):
      dir2Files.append(os.path.basename(file))
    
    for file in dir1Files:
      if not dir2Files.count(file):
        return False

    return True
    

  fsLogger.setLogLevel(LogLevel.debug)
  
  workingDir = globalConfig.getParameter("netborder.smartUpdate.Working.Directory")
  initialCfgDir = os.path.join('test-output', 'test-initial-config')
  finalCfgDir = os.path.join('test-output', 'test-final-config')
  strToReplace = 'C:\Program Files\Netborder\Express\Gateway'
  replaceDict = {strToReplace:os.getcwd()}

  tmpDirMan = TempDirManager(initialCfgDir,
                             finalCfgDir,
                             replaceDict,
                             '.properties') 
  
  if not tmpDirMan.initialize():
    log_error(fsLogger, "Failed to initialize TempDirManager")

  if not tmpDirMan.initialConvert():
    log_error(fsLogger, "TempDirManager initial convert failed")

  #Make sure that the string was replace.
  dstDir = tmpDirMan.getDstDir()
  for file in getFilesInDir(dstDir):
    for line in open(file):
      if strToReplace in line:
        log_error("String '%s' was not replace in file '%s'" % (strToReplace,
                                                                file))

  if not directoriesCompare(tmpDirMan.getSrcDir(), tmpDirMan.getDstDir()):
    log_error(fsLogger, 
              "Initial convert failed to copy all files to destination")

  if not tmpDirMan.setTempMode(): #src = tmpDir1 dst=tmpDir2
    log_error(fsLogger, "Failed to set to Temp mode")

  tmpDir1 = tmpDirMan.getSrcDir()
  tmpDir2 = tmpDirMan.getDstDir()

  if not tmpDirMan.moveRemainingFiles(): #move to tmpDir2
    log_error(fsLogger, "Failed to move remaining files")

  if not tmpDirMan.swapDirectories(): #src = tmpDir2 dst=tmpDir1
    log_error(fsLogger, "Failed to swap directories")

  if not tmpDirMan.eraseDstDirContent():
    log_error(fsLogger, "Failed to erase destination directory content")

  if not tmpDirMan.setFinalMode(): #dst = test-final-config
    log_error(fsLogger, "Failed to set mode to final")

  if not tmpDirMan.moveRemainingFiles():
    log_error(fsLogger, "Failed to move remaining files")

  if not tmpDirMan.cleanTempDirectories():
    log_error(fsLogger, "Failed to delete temp directories")


  #Make sure temporary directories are deleted.
  if os.path.exists(tmpDir1):
    log_error(fsLogger, "Directory '%s' was not deleted" % tmpDir1)
  if os.path.exists(tmpDir2):
    log_error(fsLogger, "Directory '%s' was not deleted" % tmpDir2)

  #Make sure that finalCfgDir have the same files as initialCfgDir.
  if not directoriesCompare(initialCfgDir, finalCfgDir):
    log_error(fsLogger, 
              "Initial and final directories don't have the same files")

  #Delete files in final config directory.
  tmpDirMan.eraseDstDirContent()
## !!MIRROR FILE, DO NOT MODIFY!!
